
import os
from typing import Union
from nutree import Tree, Node
from pathlib import Path
from omoospace.exceptions import InvalidError, NotFoundError, NotIncludeError
from omoospace.types import Structure, PathLike
from omoospace.utils import format_name, is_subpath
from omoospace.validators import is_entity


class Directory():
    def __init__(
        self,
        path: Path,
        is_subspace: bool = False,
    ):
        self.path = path
        self.is_subspace = is_subspace

    def __repr__(self):
        return self.path.name


class DirectoryTree(Tree):
    def __init__(self, structure: Structure = None, search_dir: PathLike = None):
        self.root_path = None
        super().__init__()
        if structure:
            self.from_structure(structure)
        if search_dir:
            self.from_dir(search_dir)

    def from_structure(
        self,
        structure: Structure,
        parent_dir: PathLike = None
    ):
        if parent_dir:
            node = self.add_dir(parent_dir)
        else:
            node = self
        self.__from_structure(node, structure)

    def from_dir(
        self,
        search_dir: PathLike,
        recursive: bool = True
    ):
        self.root_path = Path(search_dir).resolve()
        directories = self.get_dirs(search_dir, recursive=recursive)
        for directory in directories:
            self.add_dir(directory)

    def add_dir(self, dir: PathLike):
        path = Path(dir).resolve()
        if not path.is_dir():
            raise InvalidError(path, 'directory')
        if not self.root_path:
            raise NotFoundError('root directory')
        if self.root_path == path:
            return self
        if not is_subpath(path, self.root_path):
            raise NotIncludeError(path, 'root directory')

        relpath = path.relative_to(self.root_path)
        directory_paths = [relpath, *list(relpath.parents)[:-1]]
        directory_paths.reverse()
        node = self
        for directory_path in directory_paths:
            node_path = "/"+str(directory_path.as_posix())
            directory_node = self.find(
                match=lambda node: node.path == node_path
            )
            if not directory_node:
                is_subspace = is_entity(
                    Path(self.root_path, directory_path))
                directory = Directory(
                    path=directory_path,
                    is_subspace=is_subspace
                )
                directory_node = node.add(directory)
            node = directory_node
        return node

    @classmethod
    def get_dirs(
        cls,
        search_dir: PathLike,
        recursive: bool = True
    ) -> list[Path]:
        search_path: Path = Path(search_dir).resolve()
        directories: list[Path] = []

        if recursive:
            # FIXME: replace this with Path.walk (python 3.12)
            # https://docs.python.org/3/library/pathlib.html
            for root, dirs, files in os.walk(search_path):
                for dir in dirs:
                    child = Path(root, dir).resolve()
                    directories.append(child)
        else:
            for child in search_path.iterdir():
                directories.append(child)
        return directories

    def make_dirs(self, root_dir: PathLike = None):
        if not self.root_path and not root_dir:
            raise NotFoundError('root directory')
        self.__make_dirs(
            path=Path(root_dir) or self.root_path,
            node=self
        )

    @classmethod
    def __make_dirs(
        cls,
        path: Path,
        node: Union[Tree, Node],
        is_subspace: bool = False
    ):
        if (isinstance(node, Node)):
            dirpath = Path(path, node.path.removeprefix("/")).resolve()
            is_subspace = node.data.is_subspace
        else:
            dirpath = path.resolve()

        dirpath.mkdir(parents=True, exist_ok=True)
        if is_subspace:
            with Path(dirpath, "Subspace.yml").open('w') as file:
                pass

        for sub_node in node:
            cls.__make_dirs(path, node=sub_node)

    def __from_structure(
        self,
        node: Union[Tree, Node],
        structure: Structure,
    ):
        for key in structure.keys():
            parent_path = node.path if isinstance(node, Node) else "."
            directory_name = format_name(key)
            directory_path = Path(parent_path, directory_name)
            node_path = "/"+str(directory_path.as_posix())
            directory_node = self.find(
                match=lambda node: node.path == node_path
            )
            if not directory_node:
                is_subspace = key[0] == "*" \
                    or "Void" in directory_name.split("_")
                directory = Directory(
                    path=directory_path,
                    is_subspace=is_subspace
                )
                directory_node = node.add(directory)
            sub_structure = structure[key]
            if sub_structure:
                self.__from_structure(
                    node=directory_node,
                    structure=sub_structure,
                )




from pathlib import Path
import shutil
from typing import TypedDict, Union, get_args
from zipfile import ZipFile

from omoospace.directory import DirectoryTree
from omoospace.exceptions import CreateFailed, EmptyError, ExistsError, InvalidError, MoveFailed, NotFoundError, NotIncludeError
from omoospace.package import Package
from omoospace.subspace import Subspace, SubspaceTree, get_entities, get_route
from omoospace.common import ProfileContainer, ProfileItem, ProfileItemList, yaml
from omoospace.utils import format_name, remove_duplicates, reveal_in_explorer, is_subpath, copy_to
from omoospace.types import Entity, Item, PathLike, Route, Structure
from omoospace.validators import is_email, is_url, is_version


class Creator(ProfileItem):
    """Creator

    Attributes:
        email (str): Creator email
        name (str): Creator name
        website (str, optional): Creator website. 
        role (str, optional): Creator role in omoospace.
    """
    name: str
    email: str
    role: str
    website: str
    _item_list_key = 'creators'
    _item_id_key = 'email'


class Plugin(TypedDict):
    name: str
    version: str


class Software(ProfileItem):
    """Software

    Attributes:
        name (str): Software name
        version (str): Software version.
    """
    name: str
    version: str
    _item_list_key = 'softwares'
    _item_id_key = 'name'

    @property
    def plugins(self) -> list[Plugin]:
        """list[Plugin]: Software plugin list."""
        plugins = self._get_data('plugins') or []
        # remove plugin that is not dict
        plugins = [plugin for plugin in plugins
                   if isinstance(plugin, dict)]

        # remove plugin that key value is None
        plugins = [plugin for plugin in plugins
                   if plugin.get('name')]

        # remove duplicates in key
        plugins = remove_duplicates(plugins, 'name')

        # reassign modified data
        self._set_data('plugins', plugins)

        return plugins

    def set_plugin(self, name: str, version: str):
        """Add or change plugin.

        Args:
            name (str): Plugin name.
            version (str): Plugin version.
        """
        plugins = self.plugins
        for i, plugin in enumerate(plugins):
            if plugin['name'] == name:
                plugins[i] == {"name": name, "verison": version}
                self._set_data('plugins', plugins)
                return

        plugins.append({"name": name, "verison": version})
        self._set_data('plugins', plugins)


class Work(ProfileItem):
    """Work

    Attributes:
        name (str): Work name.
        description (str): Work description.
    """
    name: str
    description: str
    _item_list_key = 'works'
    _item_id_key = 'name'

    @property
    def items(self) -> list[str]:
        """list[str]: Work item list."""
        items = self._get_data('items')

        # to remove invaild items
        item_paths = self._container._to_work_item_paths(items)
        vaild_items = self._container._to_work_items(item_paths)

        self._set_data('items', vaild_items)
        return vaild_items

    def add_item(self, *items: PathLike):
        """Add item to this work.

        Args:
            *items (PathLike): Item paths.
        """
        items = self._container._to_work_items(items)
        # remove duplicates
        self._set_data('items', list(dict.fromkeys([*self.items, *items])))

    def set_items(self, *items: PathLike):
        """Set items to this work. Replace current itmes.

        Args:
            *items (PathLike): Item paths.
        """
        items = self._container._to_work_items(items)
        self._set_data('items', items)


class OmoospaceStructure(TypedDict):
    Contents: dict
    SourceFiles: dict


class OmoospaceTree(DirectoryTree):

    MAIN_DIRS = [
        "Contents",
        "ExternalData",
        "SourceFiles",
        "References",
        "StagedData"
    ]

    def __init__(self, structure: Structure = None) -> None:
        self.structure = structure or {}
        for dirname in self.MAIN_DIRS:
            if self.structure.get(dirname) is None:
                self.structure[dirname] = None
        super().__init__(structure=self.structure)


class Omoospace(ProfileContainer):
    """The class of omoospace.

    An omoospace class instance is always refer to a existed omoospace directory, not dummy. 

    Usage:
    ```python
    omoos = Omoospace('path/to/omoospace')
    print(omoos.root_path)
    # >>> path/to/omoospace
    ```

    Attributes:
        name (str): Omoospace name.
        description (str): Omoospace description.
    """

    name: str
    description: str

    def __init__(self, detect_dir: PathLike):
        """Initialize from existed omoospace.

        Args:
            detect_dir (PathLike): The start directory for detecting omoospace. It could be the subdirectories of omoospace.

        Attributes:
            name (str): Omoospace's name.
            description (str): Omoospace's description.

        Raises:
            NotFoundError: No omoospace detected.
        """
        detect_path = Path(detect_dir).resolve()
        detect_path_parents = [detect_path, *detect_path.parents]
        omoos_path = None
        for detect_path_parent in detect_path_parents:
            omoos_profile_path = Path(detect_path_parent, 'Omoospace.yml')
            if omoos_profile_path.exists():
                omoos_path = detect_path_parent
                break
        if not omoos_path:
            raise NotFoundError("omoospace", detect_dir)

        # path assignment
        self._root_path = omoos_path
        self._sourcefiles_path = Path(self.root_path, 'SourceFiles').resolve()
        self._contents_path = Path(self.root_path, 'Contents').resolve()
        self._externaldata_path = Path(
            self.root_path, 'ExternalData').resolve()
        self._stageddata_path = Path(self.root_path, 'StagedData').resolve()
        self._references_path = Path(self.root_path, 'References').resolve()
        self._profile_path = Path(self.root_path, "Omoospace.yml").resolve()

    def __check_parent_dir(self, parent_dir: PathLike):
        parent_path = Path(parent_dir).resolve(
        ) if parent_dir else self.sourcefiles_path

        # Check if is valid directory
        if not parent_path.is_dir():
            raise ExistsError(parent_dir)

        # Check if is in SourceFiles
        if not is_subpath(parent_path, self.sourcefiles_path, or_equal=True):
            raise NotIncludeError(parent_dir, "SourceFiles")

        return parent_path

    def _to_work_items(self, items: list[PathLike]) -> list[str]:

        items = [Path(item).resolve() for item in items]
        items = list(filter(self.is_contents_item, items))
        items = [str(item.relative_to(self.contents_path).as_posix())
                 for item in items]
        return list(dict.fromkeys(items))

    def _to_work_item_paths(self, items: list[str]) -> list[Path]:

        items = [Path(self.contents_path, item).resolve() for item in items]
        items = list(filter(self.is_contents_item, items))

        return list(dict.fromkeys(items))

    @property
    def root_path(self) -> Path:
        """Path: Root path."""
        return self._root_path

    @property
    def sourcefiles_path(self) -> Path:
        """Path: SourceFiles directory path."""
        return self._sourcefiles_path

    @property
    def contents_path(self) -> Path:
        """Path: Contents directory path."""
        return self._contents_path

    @property
    def externaldata_path(self) -> Path:
        """Path: ExternalData directory path."""
        return self._externaldata_path

    @property
    def stageddata_path(self) -> Path:
        """Path: StagedData directory path."""
        return self._stageddata_path

    @property
    def references_path(self) -> Path:
        """Path: Reference directory path."""
        return self._references_path

    @property
    def profile_path(self) -> Path:
        """Path: Omoospace.yml file path."""
        return self._profile_path

    @property
    def creators(self) -> list[Creator]:
        """list[Creator]: Creator List"""
        return self._get_item_list(Creator)

    @property
    def softwares(self) -> list[Software]:
        """list[Software]: Software List"""
        return self._get_item_list(Software)

    @property
    def works(self) -> list[Work]:
        """list[Work]: Work List"""
        return self._get_item_list(Work)

    @property
    def entities(self) -> list[Entity]:
        """list[Entity]: the subspace entities in the sourcefiles directory."""
        return get_entities(search_dir=self.sourcefiles_path)

    @property
    def subspace_tree(self) -> SubspaceTree:
        """SubspaceTree: The subspace tree of this omoospace."""
        return SubspaceTree(self.sourcefiles_path)

    @property
    def subspace_tree_dict(self) -> dict:
        """dict: The tree dict of subspace tree."""
        return self.subspace_tree.to_dict()

    @property
    def directory_tree(self,) -> DirectoryTree:
        """DirectoryTree: The directory tree of this omoospace."""
        return DirectoryTree(search_dir=self.root_path)

    @property
    def imported_packages(self) -> list[Package]:
        """list[Package]: The imported package list of this omoospace."""
        subdirs = [subdir for subdir in self.externaldata_path.iterdir()
                   if subdir.is_dir()]

        # Collect all Packages in ExternalData
        packages = []
        for subdir in subdirs:
            profile_file_path = Path(subdir, 'Package.yml').resolve()
            if profile_file_path.is_file():
                packages.append(Package(subdir))

        return packages

    def is_contents_item(self, path: Path) -> bool:
        """Check if path is contents item

        Args:
            path (Path): Input

        Returns:
            bool: result
        """
        exists = path.exists()
        in_contents = is_subpath(path, self.contents_path)
        return exists and in_contents

    def is_omoospace_item(self, path: Path) -> bool:
        """Check if path is this omoospace item

        Args:
            path (Path): Input

        Returns:
            bool: result
        """
        exists = path.exists()
        in_omoospace = is_subpath(path, self.root_path)
        not_in_stagedata = not is_subpath(
            path, self.stageddata_path, or_equal=True)

        not_profile_file = \
            'Omoospace.yml' != path.name \
            and 'Package.yml' != path.name \
            and 'Subspace.yml' not in path.name

        return exists and in_omoospace and not_in_stagedata and not_profile_file

    def get_subspace(self, identify: Union[Route, PathLike]) -> Subspace:
        """Get subspace object by route or entity.

        Args:
            identify (Union[Route, PathLike]): Could be route or entity path

        Raises:
            TypeError: Not vaild identify type

        Returns:
            Subspace: The wanted subspace.
        """
        if isinstance(identify, get_args(Route)):
            route = identify
            return self.subspace_tree.get(route)
        elif isinstance(identify, get_args(PathLike)):
            route = get_route(identify)
            return self.subspace_tree.get(route)
        else:
            raise TypeError

    def add_subspace(
        self,
        name: str,
        parent_dir: PathLike = None,
        description: str = None,
        reveal_in_explorer: bool = True,
        collect_entities: bool = True
    ) -> Subspace:
        """Add subspace to this omoospace.

        Args:
            name (str): Subspace name
            parent_dir (PathLike, optional): Add subspace to which directory. Defaults to None.
            description (str): str = Subspace description. Defaults to None.
            reveal_in_explorer (bool, optional): Whether open directory after or not. Defaults to True.
            collect_entities (bool, optional): Whether collect relative entities into it or not. Defaults to True.

        Raises:
            ExistsError: Target path already has one.
            NotIncludeError: Parent directory is outside SourceFiles.
            MoveFailed: Cannot move files.
            CreateFailed: Cannot create directory.

        Returns:
            Subspace: New added subspace.
        """
        parent_path = self.__check_parent_dir(parent_dir)

        subs_dirname = format_name(name)
        subs_path = Path(parent_path, subs_dirname).resolve()
        subs_profile_path = Path(subs_path, 'Subspace.yml')

        # Check if is exists
        if subs_path.is_dir():
            raise ExistsError(subs_path)

        subs_profile = {
            "name": name,
            "description": description
        }

        # create directory
        try:
            subs_path.mkdir(parents=True, exist_ok=True)
            with subs_profile_path.open('w', encoding='utf-8') as file:
                yaml.dump(subs_profile, file)
        except Exception as err:
            raise CreateFailed("subspace directory")

        # collect entities that matched.
        if (collect_entities):
            try:
                entities = get_entities(
                    parent_path, recursive=False)

                def is_match(entity: Entity):
                    entity_name = entity.stem
                    not_itself = entity_name != subs_dirname

                    is_match = False
                    entity_node_names = format_name(entity_name).split('_')
                    subs_node_names = subs_dirname.split('_')
                    for i in range(len(subs_node_names)):
                        subs_suffix = '_'.join(subs_node_names[i:])
                        entity_prefix = '_'.join(
                            entity_node_names[:len(subs_node_names) - i])
                        if subs_suffix == entity_prefix:
                            is_match = True
                            break
                    return not_itself and is_match

                # Remove entity that not match the name
                entities = list(filter(is_match, entities))

                for entity in entities:
                    shutil.move(
                        entity,
                        Path(subs_path, entity.name).resolve()
                    )
            except:
                raise MoveFailed('entities')

        if reveal_in_explorer:
            reveal_in_explorer(subs_path)

        return self.get_subspace(subs_path)

    def add_process(
        self,
        *processes: str,
        parent_dir: PathLike = None,
        add_sequence_number=True,
        reveal_in_explorer: bool = True
    ):
        """Add process to this omoospace.

        Args:
            *processes (str): a name list of processes, or single process name.
            parent_dir (PathLike, optional): Add process to which directory.. Defaults to None.
            add_sequence_number (str): Add sequence number to name. Defaults to True
            reveal_in_explorer (bool, optional): Whether open directory after or not. Defaults to True.

        Raises:
            ExistsError: Target path already has one.
            NotIncludeError: Parent directory is outside SourceFiles.
            CreateFailed: Cannot create directory.
        """
        parent_path = self.__check_parent_dir(parent_dir)

        process_structure = {}
        for i, p in enumerate(processes):
            dirname = format_name(p)
            if add_sequence_number and len(processes) > 1:
                dirname = '%03d-%s' % (i+1, dirname)
            process_structure[dirname] = None

        try:
            # TODO: may not need directory any more.
            DirectoryTree(process_structure).make_dirs(parent_path)
        except Exception as err:
            raise CreateFailed("process directory")

        if reveal_in_explorer:
            reveal_in_explorer(parent_path)

    def get_creator(self, email: str) -> Creator:
        """Get Creator by email."""
        return ProfileItemList(self.creators).find('email', email)

    def get_software(self, name: str) -> Software:
        """Get Software by name."""
        return ProfileItemList(self.softwares).find('name', name)

    def get_work(self, name: str) -> Work:
        """Get Work by name."""
        return ProfileItemList(self.works).find('name', name)

    def add_creator(
        self,
        email: str,
        name: str,
        website: str = None,
        role: str = None,
    ) -> Creator:
        """Add creator to this omoospace.

        Args:
            email (str): Creator email
            name (str): Creator name
            website (str, optional): Creator website. Defaults to None.
            role (str, optional): Creator role in omoospace. Defaults to None.

        Raises:
            InvalidError: Invalid email.
            InvalidError: Invalid website.

        Returns:
            Creator: New added creator.
        """
        if not is_email(email):
            raise InvalidError(email, 'email')

        if website and (not is_url(website)):
            raise InvalidError(email, 'url')

        creator = Creator({
            'email': email,
            'name': name,
            'website': website,
            'role': role
        }, container=self)

        self._set_profile_data(creator)

        return creator

    def add_software(
        self,
        name: str,
        version: str,
        plugins: list[Plugin] = None,
    ) -> Software:
        """Add software to this omoospace.

        Args:
            name (str): Software name
            version (str): Software version.
            plugins (list[Plugin], optional): Software plugins. Defaults to None.

        Raises:
            InvalidError: Invalid version.
            InvalidError: Invalid plugin.

        Returns:
            Software: New added software.
        """

        if not is_version(version):
            raise InvalidError(version, 'version')

        for plugin in plugins or []:
            if (not plugin.get('name')) or (not plugin.get('version')):
                raise InvalidError(plugin, 'plugin')

        software = Software({
            'name': name,
            'version': version,
            'plugins': plugins
        }, container=self)

        self._set_profile_data(software)

        return software

    def add_work(
        self,
        *items: PathLike,
        name: str = None,
        description: str = None
    ) -> Work:
        """Add work to this omoospace.

        Args:
            *items (list[PathLike]): Work items.
            name (str): Work name. Defaults to None.
            description (str, optional): Work description. Defaults to None.

        Returns:
            Work: New added work.
        """

        items = self._to_work_items(items)
        name = name or format_name(items[0].split('/')[-1])

        work = Work({
            "name": name,
            "description": description,
            "items": items
        }, container=self)

        self._set_profile_data(work)

        return work

    def export_package(
        self,
        *items: PathLike,
        name: str = None,
        export_dir: PathLike = '.',
        description: str = None,
        version: str = '0.1.0',
        reveal_in_explorer: bool = True,
        overwrite_existing: bool = True
    ) -> Package:
        """Export a package.

        Args:
            *items (PathLike): Items to export.
            name (str, optional): Package name. Defaults to None.
            export_dir (PathLike, optional): Package export directory. Defaults to None.
            description (str, optional): Package description. Defaults to None.
            reveal_in_explorer (bool, optional): Whether open directory after or not. Defaults to True.
            overwrite_existing (bool, optional): Whether overwrite existing or not. Defaults to True.

        Raises:
            ExistsError: Target path already exists.
            NotFoundError: No vaild item found.

        Returns:
            Package: New exported package.
        """
        name = name or self.name
        pkg_dirname = format_name(name)
        pkg_path = Path(export_dir, pkg_dirname).resolve()

        # Check if package dir exists
        if (pkg_path.is_dir()):
            if (overwrite_existing):
                # TODO: increase version
                shutil.rmtree(pkg_path, ignore_errors=True)
            else:
                raise ExistsError('package', pkg_path)

        items: list[Path] = [Path(item).resolve() for item in items]
        items = list(filter(self.is_omoospace_item, items))

        # Check if is enouph items
        if (len(items) == 0):
            raise NotFoundError('vaild item')

        pkg_profile = {
            "name": name,
            "version": version,
            "description": description,
            "creators": self.creators,
        }

        try:
            pkg_path.mkdir(parents=True, exist_ok=True)
            pkg_profile_path = Path(pkg_path, 'Package.yml')

            with pkg_profile_path.open('w', encoding='utf-8') as file:
                yaml.dump(pkg_profile, file)

            for i in range(len(items)):
                item = items[i]
                item_relpath = item.relative_to(self.root_path)
                copy_to(item, Path(pkg_path, item_relpath))

        except Exception as err:
            # Delete all if failed.
            shutil.rmtree(pkg_path, ignore_errors=True)
            raise Exception("Fail to export Package", err)

        if reveal_in_explorer:
            reveal_in_explorer(pkg_path)

        return Package(pkg_path)

    def import_package(
        self,
        package_path: PathLike,
        reveal_in_explorer: bool = True,
        overwrite_existing: bool = True
    ):
        """Imports the package into the ExternalData directory.

        Args:
            package_path (PathLike): Package directory.
            reveal_in_explorer (bool, optional): Whether open directory after or not. Defaults to True.
            overwrite_existing (bool, optional): Whether overwrite existing or not. Defaults to True.

        Raises:
            ExistsError: Target path already exists.
        """
        # get package form import directory
        package_path = Path(package_path).resolve()
        package = Package(package_path)

        # check if destination directory exists
        pkg_path = Path(
            self.root_path, 'ExternalData', format_name(package.name))
        if (pkg_path.is_dir()):
            if (overwrite_existing):
                shutil.rmtree(pkg_path, ignore_errors=True)
            else:
                raise ExistsError('package', pkg_path)

        if (package_path.suffix == ".zip"):
            with ZipFile(package_path, 'r') as zip:
                zip.extractall(pkg_path)
        else:
            copy_to(package_path, pkg_path)

        if reveal_in_explorer:
            reveal_in_explorer(pkg_path)


def create_omoospace(
    name: str,
    root_dir: PathLike = '.',
    description: str = None,
    structure: OmoospaceStructure = None,
    reveal_in_explorer: bool = True
) -> Omoospace:
    """Create an omoospace.

    Args:
        name (str): Omoospace name
        root_dir (PathLike, optional): Add omoospace to which directory. Defaults to '.'.
        description (str, optional): Omoospace description. Defaults to None.
        structure (OmoospaceStructure, optional): Omoospace structure. Defaults to None.
        reveal_in_explorer (bool, optional): Whether open directory after or not. Defaults to True.

    Raises:
        ExistsError: Target path already exists.
        CreateFailed: Fail to create directories.

    Returns:
        Omoospace: New created omoospace.
    """
    omoos_dirname = format_name(name)
    omoos_path = Path(root_dir, omoos_dirname).resolve()

    # Check if omoospace exists
    if (omoos_path.is_dir()):
        raise ExistsError('omoospace', root_dir)

    # write omoospace profile to yml
    omoos_profile = {
        "name": name,  # can be any name style
        "description": description,
    }

    try:
        # create dirs
        OmoospaceTree(structure).make_dirs(omoos_path)
        omoos_profile_path = Path(omoos_path, 'Omoospace.yml')
        with omoos_profile_path.open('w', encoding='utf-8') as file:
            yaml.dump(omoos_profile, file)

    except Exception as err:
        shutil.rmtree(omoos_path, ignore_errors=True)
        raise CreateFailed("omoospace directories")

    if reveal_in_explorer:
        reveal_in_explorer(omoos_path)

    return Omoospace(omoos_path)



from pathlib import Path
import re


def is_number(string):
    pattern = r'^-?\d+(?:\.\d+)?$'
    return bool(re.match(pattern, string))


def is_email(string):
    pattern = r'\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,}\b'
    return bool(re.match(pattern, string))


def is_url(string):
    pattern = re.compile(
        r'http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\\(\\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+'
    )
    return bool(re.match(pattern, string))


def is_version(string: str) -> bool:
    pattern = r'^v?\d+(\.\d+)+$|^v\d+$'
    return bool(re.match(pattern, string))


def is_autosave(string: str) -> bool:
    pattern = r'auto[-_\s]?save'
    return bool(re.match(pattern, string, re.IGNORECASE))

def is_buckup(string: str) -> bool:
    pattern = r"^bak\d*$|^backup$"
    return bool(re.match(pattern, string, re.IGNORECASE))

def is_recovered(string: str) -> bool:
    pattern = r"^recovered$"
    return bool(re.match(pattern, string, re.IGNORECASE))

def is_entity(path: str) -> bool:
    path: Path = Path(path).resolve()
    if path.is_dir():
        is_subspace = Path(path, 'Subspace.yml').is_file()
        is_void = 'Void' in path.name.split("_")
        return is_subspace or is_void
    elif path.is_file():
        not_marker = 'Subspace.yml' not in path.name
        return not_marker
    else:
        return False



import re
import os
from pathlib import Path
import shutil
from pypinyin import lazy_pinyin
from omoospace.types import PathLike
from omoospace.validators import is_autosave, is_number, is_version, is_recovered, is_buckup
import pyperclip


def format_name(string: str):
    def is_semantic(string: str):
        return (not is_number(string)) \
            and (not is_version(string)) \
            and (not is_autosave(string)) \
            and (not is_recovered(string)) \
            and (not is_buckup(string))

    # get rid of no semantic string parts.
    string = " ".join([s for s in string.split(' ')
                       if is_semantic(s)])
    string = "_".join([s for s in string.split('_')
                       if is_semantic(s)])
    string = ".".join([s for s in string.split('.')
                       if is_semantic(s)])

    # [A-Za-z0-9_-] only
    string = re.sub(r'[^\w-]', ' ', string)

    # Chinese character to pinyin
    string = ' '.join([
        string
        for string in string.split()
        for string in lazy_pinyin(string)
    ])

    # to PascalCase
    string = ' '.join([
        string.title() if string.islower() else string
        for string in string.split()
    ])
    string = '_'.join([
        string.title() if string.islower() else string
        for string in string.split("_") if string != ''
    ])

    string = string.replace(" ", "")
    return string


def remove_duplicates(list, key):
    seen = set()
    new_list = []
    for d in list:
        if d[key] not in seen:
            seen.add(d[key])
            new_list.append(d)
    return new_list


def reveal_in_explorer(dir: PathLike):
    """Open the directory in file exploarer

    Args:
        dir (PathLike): The directory want to open
    """
    try:
        os.startfile(Path(dir).resolve())
    except Exception as err:
        print("Fail to reveal", err)


def is_subpath(child: PathLike, parent: PathLike, or_equal=False) -> bool:
    """Return True if child is a subpath of parent .

    Args:
        child (PathLike): Child path
        parent (PathLike): Parent path
        or_equal (bool, optional): [description]. Defaults to False.

    Returns:
        bool: Result.
    """
    parent = Path(parent).resolve()
    child = Path(child).resolve()
    is_subpath = parent in child.parents
    is_equal = parent == child
    if or_equal:
        return is_subpath or is_equal
    else:
        return is_subpath


def copy_to(src: PathLike, dst: PathLike):
    """Copies the contents form src to dst .

    Args:
        src (Path): [description]
        dst (Path): [description]
    """
    src = Path(src).resolve()
    dst = Path(dst).resolve()
    dst.parent.mkdir(parents=True, exist_ok=True)
    if src.is_dir():
        shutil.copytree(src, dst)
    else:
        shutil.copy(src, dst)


def copy_to_clipboard(string: str):
    """Copy to clipboard

    Args:
        string (str): The string want to be copyed.
    """
    if not pyperclip.is_available():
        print("Clipboard unavailable.")
        return

    pyperclip.copy(string)


def rm_children(dir: PathLike):
    dirpath = Path(dir).resolve()
    for child in dirpath.iterdir():
        if child.is_file():
            child.unlink()
        else:
            shutil.rmtree(child, ignore_errors=True)



from omoospace.omoospace import (
    Omoospace,
    Creator,
    Software,
    Work,
    create_omoospace
)

from omoospace.subspace import (
    SubspaceTree,
    SubspaceType,
    Subspace,
    get_route,
    get_route_str
)

from omoospace.package import (
    Package
)

from omoospace.utils import (
    format_name,
    copy_to_clipboard,
    reveal_in_explorer
)

from omoospace.types import *

from omoospace.validators import *



from pathlib import Path

from omoospace.directory import DirectoryTree, Structure
from omoospace.exceptions import InvalidError
from omoospace.types import Creator, OmoospaceStructure
from omoospace.omoospace import Omoospace, OmoospaceTree
from omoospace.common import console, yaml


class OmoospaceTemplate():
    def __init__(
        self,
        name: str = None,
        description: str = None,
        structure: OmoospaceStructure = None
    ) -> None:
        self.name = name
        self.description = description
        self.structure = structure or {}
        self.tree = OmoospaceTree(structure=structure)


class ProcessTemplate():
    def __init__(
        self,
        name: str,
        description: str,
        structure: Structure = None
    ) -> None:
        self.name = name
        self.description = description
        self.structure = structure or {}
        self.tree = DirectoryTree(structure=structure)


class Setting():
    working_omoospace: str
    working_directory: str
    recent_omoospaces: list[str]
    omoospace_templates: list[OmoospaceTemplate]
    process_templates: list[ProcessTemplate]
    registered_creators: list[Creator]
    software_choices: list[str]
    role_choices: list[str]

    SETTING_PATH = "~/.omoospace/setting.yml"

    DEFAULT_SETTING = {
        "working_omoospace": None,
        "working_directory": None,
        "recent_omoospaces": None,
        "omoospace_templates": [
            {
                "name": "3D Asset",
                "description": "For 3D asset creation",
                "structure": {
                    "Contents": {
                        'Images': None,
                        'Materials': None,
                        'Models': None,
                        'Renders': None
                    },
                    "SourceFiles": {
                        '*AssetName': None,
                        'Void': None
                    }
                }
            },
            {
                "name": "CG Film",
                "description": "For short CG film producton",
                "structure": {
                    "Contents": {
                        'Audios': None,
                        'Dynamics': None,
                        'Images': None,
                        'Materials': None,
                        'Models': None,
                        'Renders': None,
                        'Videos': None,
                    },
                    "SourceFiles": {
                        '*FilmName': None,
                        'Void': None
                    }
                }
            }
        ],
        "process_templates": [
            {
                "name": "Asset Creation",
                "description": "Asset Creation Process",
                "structure": {
                    "001-Modeling": None,
                    "002-Texturing": None,
                    "003-Shading": None
                }
            },
            {
                "name": "Film Production",
                "description": "Film Production Process",
                "structure": {
                    "001-PreProduction": None,
                    "002-Production": None,
                    "003-PostProduction": None
                }
            }
        ],
        "registered_creators": [],
        "role_choices": [
            "Owner",
            "Menber",
        ],
        "software_choices": [
            "Blender",
            "Cinema4D",
            "Zbrush"
        ]
    }

    def __read_setting_file(self):
        setting: dict = {}
        setting_filepath = Path(self.SETTING_PATH).expanduser().resolve()
        if setting_filepath.exists():
            with setting_filepath.open('r', encoding='utf-8') as file:
                # aviod empty or invalid ifle
                setting = yaml.load(file) or {}
        return setting

    def __write_setting_file(self, setting):
        setting_filepath = Path(self.SETTING_PATH).expanduser().resolve()
        with setting_filepath.open('w', encoding='utf-8') as file:
            yaml.dump(setting, file)

    def __getattr__(self, key):
        if key not in self.__annotations__.keys():
            raise InvalidError(key, "key")
        setting = self.__read_setting_file()
        if (key == "omoospace_templates"):
            omoospace_templates = setting.get("omoospace_templates") or []
            omoospace_templates.extend(
                self.DEFAULT_SETTING["omoospace_templates"])
            return [OmoospaceTemplate(
                name=template.get('name'),
                description=template.get('description'),
                structure=template.get('structure')
            ) for template in omoospace_templates]
        elif (key == "process_templates"):
            process_templates = setting.get("process_templates") or []
            process_templates.extend(
                self.DEFAULT_SETTING["process_templates"])
            return [ProcessTemplate(
                name=template.get('name'),
                description=template.get('description'),
                structure=template.get('structure')
            ) for template in process_templates]
        elif (key == "software_choices"):
            software_choices = setting.get("software_choices") or []
            software_choices += list(
                set(self.DEFAULT_SETTING["software_choices"])
                - set(software_choices))
            return software_choices
        elif (key == "role_choices"):
            role_choices = setting.get("role_choices") or []
            role_choices += list(
                set(self.DEFAULT_SETTING["role_choices"])
                - set(role_choices))
            return role_choices
        else:
            return setting.get(key) or self.DEFAULT_SETTING.get(key)

    def __setattr__(self, key, value):
        if key not in self.__annotations__.keys():
            raise InvalidError(key, "key")
        setting = self.__read_setting_file()
        if (key == "software_choices"):
            value = list(
                set(value)
                - set(self.DEFAULT_SETTING["software_choices"]))
        if (key == "role_choices"):
            value = list(
                set(value)
                - set(self.DEFAULT_SETTING["role_choices"]))
        setting[key] = value
        self.__write_setting_file(setting)


def to_omoospace(detect_path: str):
    omoospace = Omoospace(detect_path)
    recent_list = Setting().recent_omoospaces or []
    omoospace_dir = str(omoospace.root_path)
    # remove old record in settings
    try:
        recent_list.remove(omoospace_dir)
    except:
        pass
    recent_list.append(omoospace_dir)
    Setting().recent_omoospaces = recent_list
    Setting().working_omoospace = omoospace_dir
    console.print("Current working at 🛠️ [bright]%s (%s)[/bright]." %
                  (omoospace.name, omoospace_dir))
    return omoospace





import os
from pathlib import Path
from typing import TypedDict
from nutree import Tree, Node
from enum import Enum

from omoospace.exceptions import InvalidError, NotFoundError
from omoospace.types import Entity, PathLike, Route
from omoospace.common import yaml
from omoospace.utils import format_name
from omoospace.validators import is_entity


class SubspaceType(Enum):
    DIRECTORY = "directory"
    FILE = "file"
    PHANTOM = "phantom"
    DUMMY = "dummy"


class SubspaceData():
    def __init__(self, node_name: str):
        self.node_name = node_name
        self.entities: list[Entity] = []

    def __repr__(self):
        return self.node_name


class Subspace():
    """Subspace object.

    Attributes:
        name (str): Subsapce display name.
        description (str): Subsapce description.
    """
    name: str
    description: str

    def __init__(self, node: Node):
        self._node = node

    def __repr__(self):
        return self.node_name

    def __read_profile_file(self):
        if self.profile_path.is_file():
            with self.profile_path.open('r', encoding='utf-8') as file:
                # aviod empty or invalid file
                profile = yaml.load(file) or {}
            return profile
        else:
            return {}

    def __write_profile_file(self, profile):
        with self.profile_path.open('w', encoding='utf-8') as file:
            yaml.dump(profile, file)

    def __getattr__(self, name):
        if name in self.__annotations__.keys():
            if self.type == SubspaceType.DUMMY:
                return None
            else:
                profile = self.__read_profile_file()
                return profile.get(name)
        else:
            return object.__getattribute__(self, name)

    def __setattr__(self, name, value):
        if name in self.__annotations__.keys():
            if self.type == SubspaceType.DUMMY:
                pass
            else:
                profile = self.__read_profile_file()
                profile[name] = value
                self.__write_profile_file(profile)
        else:
            object.__setattr__(self, name, value)

    @property
    def root_path(self) -> Path:
        """Path: Subspace root path (directory subspace only)."""
        for entity in self.entities:
            if entity.is_dir():
                return entity
        return None

    @property
    def profile_path(self) -> Path:
        """Path: Subspace profile file path."""
        if self.type == SubspaceType.DIRECTORY:
            return Path(self.root_path, 'Subspace.yml').resolve()
        elif self.type != SubspaceType.DUMMY:
            entity = self.entities[0]
            node = self
            while True:
                parent = node.parent
                if not parent:
                    dir_parent = None
                    break
                elif parent.type == SubspaceType.DIRECTORY:
                    dir_parent = parent
                    break
                node = parent
            sub_route = self.route[len(
                dir_parent.route):] if dir_parent else self.route
            profile_filename = '%s.Subspace.yml' % "_".join(sub_route)
            return Path(entity.parent, profile_filename).resolve()
        else:
            return None

    @property
    def type(self) -> SubspaceType:
        """SubspaceType: Subspace type."""
        if len(self.entities) != 0:
            if len(self.endpoint_entities) != 0:
                if self.root_path:
                    return SubspaceType.DIRECTORY
                else:
                    return SubspaceType.FILE
            else:
                return SubspaceType.PHANTOM
        else:
            return SubspaceType.DUMMY

    @property
    def endpoint_entities(self) -> list[Entity]:
        """list[Entity]: Subspace endpoint entities."""
        phantom_entities = []
        for entity in self.entities:
            node_names = format_name(entity.stem).split("_")
            if self.node_name == node_names[-1]:
                phantom_entities.append(entity)
        return phantom_entities

    @property
    def entities(self) -> list[Entity]:
        """list[Entity]: Subspace entities."""
        entities = self._node.data.entities

        def is_vaild(entity: Entity):
            exists = entity.exists()
            return exists

        # real file/directory only. remove those not exists.
        entities = list(filter(is_vaild, entities))
        self._node.data.entities = entities

        return entities

    @property
    def node_name(self) -> str:
        """str: Subspace node name."""
        return self._node.data.node_name

    @property
    def parent(self) -> "Subspace":
        """Subspace: Parent subspace."""
        node = self._node.parent
        return Subspace(node) if node else None

    @property
    def route(self) -> Route:
        """Route: Substance route."""
        path = self._node.path
        return path.split("/")[1:]

    @property
    def children(self) -> list["Subspace"]:
        """list["Subspace"]: Children subspace."""
        nodes = self._node.children
        return [Subspace(node) for node in nodes]

    def _add_entity(self, *entities: Entity):
        for entity in entities:
            if entity not in self.entities:
                self._node.data.entities.append(entity)

    def add(self, node_name: str) -> "Subspace":
        """Add dummy subpace to it.

        Args:
            node_name (str): New add subspace node name.

        Returns:
            Subspace: New added dummy subspace.
        """
        node = self._node.add(SubspaceData(node_name))
        return Subspace(node)


class SubspaceTree():
    """Subspace tree object."""

    def __init__(
        self,
        path: PathLike = None
    ):
        """Initialize the subspace tree from the directory or certain entity.

        Args:
            path (PathLike, optional): Build subspace from path. Defaults to None.
        """
        self._tree = Tree()
        if path:
            path = Path(path).resolve()
            if path.is_dir():
                self.from_dir(path)
            elif path.is_file():
                self.from_entity(path)
            else:
                NotFoundError()

    def from_dir(
        self,
        search_dir: PathLike,
        recursive: bool = True
    ):
        """Build the subspace tree from directory.

        It will search all vaild entities in giving directory, then build the tree based on those entities.

        Args:
            search_dir (PathLike): The search directory.
            recursive (bool, optional): Whether recursive or not. Defaults to True.
        """
        entities = get_entities(search_dir, recursive=recursive)
        for entity in entities:
            self.from_entity(entity)

    def get(self, route: Route) -> Subspace:
        """Get subspace by route.

        Args:
            route (Route): Input.

        Returns:
            Subspace: Result.
        """
        node = self._tree.find(
            match=lambda node: node.path == "/" + "/".join(route)
        )
        return Subspace(node) if node else None

    def add(self, node_name: str) -> Subspace:
        """Add dummy subpace to tree root.

        Args:
            node_name (str): New add subspace node name.

        Returns:
            Subspace: New added dummy subspace.
        """
        node = self._tree.add(SubspaceData(node_name))
        return Subspace(node)

    def to_dict(self) -> dict:
        """Convert the tree to a common dict.

        Returns:
            dict: Result.
        """
        def mapper(node: Node, data: dict):
            data['data'] = Subspace(node)
            return data
        return self._tree.to_dict_list(mapper=mapper)

    def from_entity(self, entity_path: PathLike):
        """Build the subspace tree from one certain entity.

        The path must be real, not dummy.

        Args:
            entity_path (Entity): The giving entity.

        Raises:
            InvalidError: Invalid entity.
        """
        route_entities = get_route_entities(entity_path)

        node = self
        for i in range(len(route_entities)):
            d = route_entities[i]
            route = [d.get('node_name') for d in route_entities[:i+1]]
            node_name = d.get('node_name')
            entities = d.get('entities')
            subspace = self.get(route)
            if not subspace:
                subspace = node.add(node_name)
            subspace._add_entity(*entities)
            node = subspace


def get_entities(
    search_dir: PathLike,
    recursive: bool = True
) -> list[Entity]:
    """Search all vaild entities in giving directory.

    Args:
        search_dir (PathLike): The search directory.
        recursive (bool, optional): Whether recursive or not. Defaults to True.
    """
    search_path: Path = Path(search_dir).resolve()
    entities: list[Entity] = []
    if recursive:
        # FIXME: replace this with Path.walk (python 3.12)
        # https://docs.python.org/3/library/pathlib.html
        for root, dirs, files in os.walk(search_path):
            for path in [*dirs, *files]:
                child = Path(root, path).resolve()
                if is_entity(child):
                    entities.append(child)
    else:
        for child in search_path.iterdir():
            if is_entity(child):
                entities.append(child)
    return entities


def get_route_entities(entity_path: PathLike) -> list:
    # Check if input entity is vaild.
    entity = Path(entity_path).resolve()
    if not is_entity(entity):
        InvalidError(entity, 'entity')

    # Get relpath parts. remove those directory that is not entity.
    entities = [entity]
    for parent in entity.parents:
        if is_entity(parent):
            entities.append(parent)
    entities.reverse()

    # Generate route.
    route_entities = []
    for entity in entities:
        route: Route = [d.get('node_name') for d in route_entities]

        # format namespace
        entity_name = format_name(entity.stem)

        # node names are the strings that splited by "_" from entity name.
        # e.g. `SQ010_SH0100.blend` has two namespace: `SQ010` and `SH0100`
        node_names: list[str] = entity_name.split('_')

        # clip matched route names as mush as possible.
        # e.g. route: ['SQ010'], enity name: `SQ010_SH0100.blend`
        # `SQ010` is the matched namespace, which will be cliped.
        for i in range(len(route)):
            route_suffix = '_'.join(route[i:])
            node_prefix = '_'.join(node_names[:len(route) - i])
            if route_suffix == node_prefix:
                node_names = node_names[len(route) - i:]
                break

        # append node_name to route with entity
        for node_name in node_names:
            route_entities.append({
                'node_name': node_name,
                'entities': [entity]
            })

        # Sometimes all namspaces are cliped. but still need to append entity
        if len(node_names) == 0:
            route_entities[-1].get('entities').append(entity)

    return route_entities


def get_route(entity_path: PathLike) -> Route:
    """Returns the subspace route of a entity.

    Super useful for creating path of source file outputs. For example:
    ```python
    route = get_route("path/to/SQ010/SQ010_AssetA.blend")
    # >>> ['SQ010','AssetA']
    ```
    
    Args:
        entity_path (PathLike): The giving entity.

    Returns:
        Route: Subspace route.
    """
    route_entities = get_route_entities(entity_path)
    route: Route = [d.get('node_name') for d in route_entities]
    return route


def get_route_str(entity_path: PathLike, *subsets: str) -> str:
    """Returns the string based on entity's subspace route.

    Super useful for creating path of source file outputs. For example:
    ```python
    string = get_route_str("path/to/SQ010/SQ010_AssetA.blend","HighRes","v001")
    # >>> SQ010_AssetA_HighRes_v001
    ```
    
    Args:
        entity_path (PathLike): The giving entity.

    Returns:
        str: Result.
    """
    route = get_route(entity_path)
    route.extend(subsets)
    return "_".join(route)



import os
from pathlib import Path
from zipfile import ZipFile


from omoospace.exceptions import NotFoundError
from omoospace.types import Item, PathLike
from omoospace.common import ProfileContainer, yaml
from omoospace.utils import is_subpath


class Package(ProfileContainer):
    """The class of omoospace package.

    A package class instance is always refer to a existed package directory, not dummy. 

    Usage:
    ```python
    pkg = Package('path/to/package')
    print(pkg.root_path)
    # >>> path/to/package
    ```
    
    Attributes:
        name (str): Package's name.
        description (str): Package's description.
        version (str): Package's version.
    """
    name: str
    description: str
    version: str

    def __init__(self, package_path: PathLike):
        """Initialize package .

        Args:
            package_path (PathLike): The package path.

        Raises:
            NotFoundError: No package detected.
        """
        package_path = Path(package_path).resolve()
        if (package_path.suffix == ".zip"):
            with ZipFile(package_path, 'r') as zip:
                try:
                    with zip.open('Package.yml') as file:
                        pass
                except:
                    raise NotFoundError("package", package_path)
        else:
            package_info_path = Path(package_path, 'Package.yml')
            if not package_info_path.exists():
                raise NotFoundError("package", package_path)

        self._root_path = package_path
        self._profile_path = Path(self.root_path, "Package.yml").resolve()

    @property
    def root_path(self) -> Path:
        """Path: Root path."""
        return self._root_path

    @property
    def profile_path(self) -> Path:
        """Path: Package.yml file path."""
        return self._profile_path

    def is_package_item(self, path: Path) -> bool:
        """Check if path is package item

        Args:
            path (Path): Input

        Returns:
            bool: result
        """
        exists = path.exists()
        in_package = is_subpath(path, self.root_path)
        not_profile_file = \
            'Omoospace.yml' != path.name \
            and 'Package.yml' != path.name \
            and 'Subspace.yml' not in path.name

        return exists and in_package and not_profile_file

    @property
    def items(self) -> list[Item]:
        """list[Item]: Package item list."""
        items: list[Item] = []
        for root, dirs, files in os.walk(self.root_path):
            for path in files:
                child = Path(root, path).resolve()
                if self.is_package_item(child):
                    items.append(child)
        return items



from pathlib import Path
from ruamel.yaml import YAML
from abc import ABC, abstractmethod
from omoospace.utils import remove_duplicates

yaml = YAML()
yaml.indent(sequence=4, offset=2)


class AttributeDict(ABC):
    @abstractmethod
    def _get_data(self, name):
        return None

    @abstractmethod
    def _set_data(self, name, value):
        pass

    def __getattr__(self, name):
        if name in self.__annotations__.keys():
            return self._get_data(name)
        else:
            return object.__getattribute__(self, name)

    def __setattr__(self, name, value):
        if name in self.__annotations__.keys():
            self._set_data(name, value)
        else:
            object.__setattr__(self, name, value)


class ProfileItem(AttributeDict):
    _item_list_key = 'items'
    _item_id_key = 'id'

    def __init__(self, dict: dict, container: 'ProfileContainer') -> None:
        self._dict = dict
        self._container = container

    def _get_data(self, name):
        return self._dict.get(name)

    def _set_data(self, name, value):
        identify = self._dict[self._item_id_key]
        self._dict[name] = value
        self._container._set_profile_data(self, identify)


class ProfileItemList():
    def __init__(self, item_list: list[ProfileItem]) -> None:
        self._item_list = item_list

    def set(
        self,
        data: ProfileItem,
        identify: str = None
    ):
        _item_id_key = data._item_id_key
        identify = identify or getattr(data, _item_id_key)

        index = self.__get_matched_index(_item_id_key, identify)

        if index >= 0:
            self._item_list[index] = data
        else:
            self._item_list.append(data)

    def __get_matched_index(
        self,
        key: str,
        value: any
    ):
        index = -1
        for i, data in enumerate(self._item_list):
            if getattr(data, key) == value:
                index = i
                break

        return index

    def to_dict_list(self):
        return [data._dict for data in self._item_list]

    def find(
        self,
        key: str,
        value: any
    ):
        index = self.__get_matched_index(key, value)

        if index >= 0:
            return self._item_list[index]
        else:
            return None


class ProfileContainer(AttributeDict):
    profile_path: Path = Path()

    def _read_profile_file(self):
        with self.profile_path.open('r', encoding='utf-8') as file:
            # aviod empty or invalid ifle
            profile = yaml.load(file) or {}
        return profile

    def _write_profile_file(self, profile):
        with self.profile_path.open('w', encoding='utf-8') as file:
            yaml.dump(profile, file)

    def _get_data(self, name):
        profile = self._read_profile_file()
        return profile.get(name)

    def _get_item_list(self, data_class):
        item_list_key = data_class._item_list_key
        item_id_key = data_class._item_id_key

        profile = self._read_profile_file()
        item_list = profile.get(item_list_key) or []

        # remove data that is not dict
        item_list = [data for data in item_list
                     if isinstance(data, dict)]

        # remove data that key value is None
        item_list = [data for data in item_list
                     if data.get(item_id_key)]

        # remove duplicates in key
        item_list = remove_duplicates(item_list, item_id_key)

        # reassign modified data
        self._set_data(item_list_key, item_list)

        # wrap dict into class
        return [data_class(data, self) for data in item_list]

    def _set_data(self, name, value):
        profile = self._read_profile_file()
        profile[name] = value
        self._write_profile_file(profile)

    def _set_profile_data(self, data: ProfileItem, identify: str = None):
        item_list_key = data._item_list_key

        item_list = ProfileItemList(getattr(self, item_list_key))
        item_list.set(data, identify)
        self._set_data(item_list_key, item_list.to_dict_list())



class NotFoundError(Exception):
    def __init__(self,  item: str = None, scope: str = None):
        item_str = '%s ' % item if item else ""
        scope_str = ' in "%s"' % scope if scope else ""
        message = 'No %sfound%s.' % (item_str, scope_str)
        super().__init__(message)


class ExistsError(Exception):
    def __init__(self,  item: str = None, scope: str = None):
        item_str = '%s ' % item if item else ""
        scope_str = ' in "%s"' % scope if scope else ""
        message = '%salready exists%s.' % (item_str, scope_str)
        super().__init__(message)


class NotIncludeError(Exception):
    def __init__(self,  item: str = None, scope: str = None):
        item_str = '%s is ' % item if item else ""
        scope_str = ' in "%s"' % scope if scope else ""
        message = '%snot include%s.' % (item_str, scope_str)
        super().__init__(message)


class InvalidError(Exception):
    def __init__(self,  item: str = None, type: str = None):
        item_str = '%s ' % item if item else ""
        type_str = ' %s' % type if type else ""
        message = '%sis invalid%s.' % (item_str, type_str)
        super().__init__(message)


class EmptyError(Exception):
    def __init__(self,  item: str = None):
        item_str = '%s ' % item if item else ""
        message = '%scannot be empty.' % item_str
        super().__init__(message)


class CreateFailed(Exception):
    def __init__(self, item: str = None):
        item_str = ' %s' % item if item else ""
        message = 'Fail to create%s.' % item_str
        super().__init__(message)

class MoveFailed(Exception):
    def __init__(self, item: str = None):
        item_str = ' %s' % item if item else ""
        message = 'Fail to move%s.' % item_str
        super().__init__(message)